/*

findx -- combine find and xargs
Copyright (C) 2007 Alexander Foken

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 51
Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

*/

/* See also: http://www.cse.ohio-state.edu/~mamrak/CIS762/pipes_lab_notes.html */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/wait.h>

extern const char usage[];
extern char findpath[];
extern char xargspath[];

const char cvsid[] = "$Id: findx.c,v 1.7 2008/05/14 09:40:05 alex Exp $";

#ifdef DEBUG
	#define debugf(fmt,...) fprintf(stderr,fmt, __VA_ARGS__ )
#else
	#define debugf(fmt,...) /* nop */
#endif

#ifdef DEBUG
static void dump_exec(char * name, char ** args)
{
	pid_t pid=getpid();
	debugf("executing '%s' with pid=%i with the arguments",name,pid);
	while (args[0]) {
		debugf(" '%s'",args[0]);
		args++;
	}
	debugf("%s","\n");
}
#else
	#define dump_exec(name,args) /* nop */
#endif

static void pipe_in(int fds[])
{
	close(fds[1]);
	close(0); /* stdin */
	dup2(fds[0],0); /* stdin is now the pipe */
	close(fds[0]);
}

static void pipe_out(int fds[])
{
	close(fds[0]);
	close(1); /* stdout */
	dup2(fds[1],1); /* stdout is now the pipe */
	close(fds[1]);
}

static char ** get_argv_xargs(int argc, char ** argv, int marker_pos)
{
	/*
		We need to insert the xargs command and two arguments: "-0" and "-r".
		In some cases, we only have two places, the "--" and the findx location ("findx -- command").
		We have to move all arguments one element up.
		
		So lat's allocate what we really need: Room for xargs, the two arguments to be inserted,
		(argc-marker_pos-1) remaining arguments, plus a final NULL pointer = argc+3+1-marker_pos-1 =argc+3-marker_pos.

		And we don't care about releasing the allocated memory, as this happens just before exec() or exit()
		in a child process.
	*/
	
	int i,j=0;
	char ** rv=calloc(argc+3-marker_pos,sizeof(char *));
	if (rv==NULL) {
		perror("calloc");
		exit(1);
	}
	
	rv[j++]=xargspath;
	rv[j++]="-0";
	rv[j++]="-r";
	for (i=marker_pos+1; i<argc; i++) {
		rv[j++]=argv[i];
	}
	rv[j++]=NULL;
	
	return rv;
}

static char ** get_argv_find(int argc, char ** argv, int marker_pos)
{
	/*
		We need to add three arguments to the find arguments: '(', ')', and '-print0'.
		In some cases, we only have two places, the "--" and the command for xargs. ("findx -tests -- echo").
		We have to move all but the first few arguments one element up.
		
		So let's allocate what we really need: Room for marker_pos + 3 elements, plus a final NULL pointer.
		
		And we don't care about releasing the allocated memory, as this happens just before exec() or exit()
		in a child process.
	*/

	int i,j,start;	
	char ** rv=calloc(4+marker_pos,sizeof(char *));
	if (rv==NULL) {
		perror("calloc");
		exit(1);
	}
	
	rv[0]=findpath;
	start=0;
	j=1;
	for (i=1; i<marker_pos; i++) {
		if (!start) {
			switch (argv[i][0]) {
				case '(' :
				case ')' :
				case ',' :
				case '!' :
					start++;
					break;
				case '-' :
					switch (argv[i][1]) {
						case 'H' :
						case 'L' :
						case 'P' :
						case 0 :
							break;
						default :
							start++;
							break;
					}
					break;
			}
			if (start) {
				rv[j++]="(";
			}
		}
		rv[j++]=argv[i];
	}
	rv[j++]=")";
	rv[j++]="-print0";
	rv[j++]=NULL;
	
	return rv;
}

static pid_t fork_and_exec(void (*setup_pipe)(int []), int fds[], char ** (*get_argv)(int,char **,int),int argc, char ** argv, int marker_pos)
{
	pid_t pid;

	fflush(NULL);
	
	pid=fork();
	if (pid==-1) {
		perror("fork");
		exit(1);
	}
	if (pid!=0) {
		debugf("forked pid=%i\n",pid);
		return pid;
	}
	
	setup_pipe(fds);
	argv=get_argv(argc,argv,marker_pos);

	dump_exec(argv[0],argv);

	execv(argv[0],argv);
	perror("exec");
	exit(1);	
}

int main(int argc, char ** argv)
{
	int i;
	int fds[2];
	pid_t dead;
	pid_t pid[2];

	for (i=1; i<argc; i++) {
		if (strcmp(argv[i],"--")==0) {
			break;
		}
	}
	if ((i==argc) || (argc<3)) {
		fprintf(stderr,"%s\nCVS ID: %s\n",usage,cvsid);
		return 1;
	}

	if (pipe(fds)==-1) {
		perror("pipe");
		return 1;
	}

	pid[0]=fork_and_exec(pipe_in,fds,get_argv_xargs,argc,argv,i);
	pid[1]=fork_and_exec(pipe_out,fds,get_argv_find,argc,argv,i);

	close(fds[0]);
	close(fds[1]);
	
	debugf("waiting for dead children %d and %d\n",pid[0],pid[1]);
	while (pid[0] || pid[1]) {	
		dead=wait(&i);
		debugf("process %d died\n",dead);
		if (dead==-1) {
			perror("wait");
			return 1;
		}
		if (dead==pid[0]) {
			pid[0]=0;
		} else if (dead==pid[1]) {
			pid[1]=0;
		}
	}
	
	return 0;
}
